@@ -449,11 +449,10 @@ class SimplifySyntaxTree_Visitor(SVisitor):
             if to_remove:
                 tree.remove_kids_by_index(*to_remove)
 
-class FilterTokens_Tranformer(STransformer):
+class FilterTokens_Visitor(SVisitor):
     def __default__(self, tree):
-        if len(tree.tail) <= 1:
-            return tree
-        return tree.__class__(tree.head, [x for x in tree.tail if is_stree(x)])
+        if len(tree.tail) > 1:
+            tree.tail = filter(is_stree, tree.tail)
 
 class TokValue(Str):
     def __new__(cls, s, type=None, line=None, column=None, pos_in_stream=None, index=None):
@@ -465,12 +464,6 @@ class TokValue(Str):
         inst.index = index
         return inst
 
-    # XXX Required to exclude 'parent'
-    def __getstate__(self):
-        return self.type, self.line, self.column, self.pos_in_stream, self.index
-    def __setstate__(self, x):
-        self.type, self.line, self.column, self.pos_in_stream, self.index = x
-
 class LexerWrapper(object):
     def __init__(self, lexer, newline_tokens_names, newline_char='\n', ignore_token_names=()):
         self.lexer = lexer
@@ -533,15 +526,25 @@ class LexerWrapper(object):
 
 class Grammar(object):
     def __init__(self, grammar, **options):
-        if hasattr(grammar, 'read'):
-            # PLY turns "a.b" into "b", so gotta get rid of the dot.
-            tab_filename = "parsetab_%s" % os.path.split(grammar.name)[1].replace('.', '_')
+        # Some, but not all file-like objects have a 'name' attribute
+        try:
             source = grammar.name
-            grammar = grammar.read()
-        else:
-            assert isinstance(grammar, StringTypes)
-            tab_filename = "parsetab_%s" % str(hash(grammar)%(2**32))
+        except AttributeError:
             source = '<string>'
+            tab_filename = "parsetab_%s" % str(hash(grammar)%(2**32))
+        else:
+            # PLY turns "a.b" into "b", so gotta get rid of the dot.
+            tab_filename = "parsetab_%s" % os.path.basename(source).replace('.', '_')
+
+        # Drain file-like objects to get their contents
+        try:
+            read = grammar.read
+        except AttributeError:
+            pass
+        else:
+            grammar = read()
+
+        assert isinstance(grammar, StringTypes)
 
         grammar_tree = grammar_parser.parse(grammar)
         if not grammar_tree:
@@ -571,8 +574,8 @@ class _Grammar(object):
         self.tab_filename = tab_filename
         self.source_name = source_name
         self.tokens = []    # for lex module
-        self.rules_to_flatten = []
-        self.rules_to_expand = []
+        self.rules_to_flatten = set()
+        self.rules_to_expand = set()
         self._newline_tokens = set()
         self._ignore_tokens = set()
         self.lexer_postproc = None
@@ -594,7 +597,13 @@ class _Grammar(object):
             ply_grammar, = ply_grammar_and_code
         else:
             ply_grammar, code = ply_grammar_and_code
-            exec(code, locals())
+
+            # prefix with newlines to get line-number count correctly (ensures tracebacks are correct)
+            src_code = '\n' * (max(code.line, 1) - 1) + code
+
+            # compiling before executing attaches source_name as filename: shown in tracebacks
+            exec_code = compile(src_code, source_name, 'exec')
+            exec(exec_code, locals())
 
         for x in ply_grammar:
             type_, (name, defin) = x.head, x.tail
@@ -642,7 +651,7 @@ class _Grammar(object):
 
         # Apply auto-filtering (remove 'punctuation' tokens)
         if self.auto_filter_tokens:
-            tree = FilterTokens_Tranformer().transform(tree)
+            FilterTokens_Visitor().visit(tree)
 
         SimplifySyntaxTree_Visitor(self.rules_to_flatten, self.rules_to_expand, self.keep_empty_trees).visit(tree)
 
@@ -721,20 +730,18 @@ class _Grammar(object):
 
                     self.tokens.append(name)
 
-                    code = ('\tt.type = self._%s_unless_toks_dict.get(t.value, %r)\n' % (name, name)
-                           +'\tfor regexp, tokname in self._%s_unless_toks_regexps:\n' % (name,)
-                           +'\t\tif regexp.match(t.value):\n'
-                           +'\t\t\tt.type = tokname\n'
-                           +'\t\t\tbreak\n'
-                           +'\treturn t')
-                    s = ('def t_%s(self, t):\n\t%s\n%s\nx = t_%s\n'
-                        %(name, repr(token_value), code, name))
-                    d = {}
-                    exec(s, d)
+                    def t_token(self, t):
+                        t.type = getattr(self, '_%s_unless_toks_dict' % (name,)).get(t.value, name)
+                        for regexp, tokname in getattr(self, '_%s_unless_toks_regexps' % (name,)):
+                            if regexp.match(t.value):
+                                t.type = tokname
+                                break
+                        return t
+                    t_token.__doc__ = token_value
 
-                    setattr(self, 't_%s'%name, d['x'].__get__(self))
-                    setattr(self, '_%s_unless_toks_dict'%name, unless_toks_dict)
-                    setattr(self, '_%s_unless_toks_regexps'%name, unless_toks_regexps)
+                    setattr(self, 't_%s' % (name,), t_token.__get__(self))
+                    setattr(self, '_%s_unless_toks_dict' % (name,), unless_toks_dict)
+                    setattr(self, '_%s_unless_toks_regexps' % (name,), unless_toks_regexps)
 
                     token_added = True
 
@@ -766,19 +773,32 @@ class _Grammar(object):
             rule_name = rule_name[len(mods):]
 
         if RuleMods.EXPAND in mods:
-            self.rules_to_expand.append( rule_name )
+            self.rules_to_expand.add( rule_name )
         elif RuleMods.FLATTEN in mods:
-            self.rules_to_flatten.append( rule_name )
+            self.rules_to_flatten.add( rule_name )
 
-        if RuleMods.EXPAND1 in mods or RuleMods.EXPAND in mods:  # EXPAND is here just for the speed-up
-            code = '\tp[0] = self.tree_class(%r, p[1:], skip_adjustments=True) if len(p)>2 else p[1]' % (rule_name,)
-        else:
-            code = '\tp[0] = self.tree_class(%r, p[1:], skip_adjustments=True)' % (rule_name,)
-        s = ('def p_%s(self, p):\n\t%r\n%s\nx = p_%s\n'
-            %(rule_name, rule_def, code, rule_name))
-        d = {}
-        exec(s, d)
-        setattr(self, 'p_%s'%rule_name, types.MethodType(d['x'], self))
+        def p_rule(self, p):
+            subtree = []
+            for child in p[1:]:
+                if isinstance(child, self.tree_class) and (
+                           (                            child.head in self.rules_to_expand )
+                        or (child.head == rule_name and child.head in self.rules_to_flatten)
+                        ):
+                    # (EXPAND | FLATTEN) & mods -> here to keep tree-depth minimal, prevents unbounded tree-depth on
+                    #                              recursive rules.
+                    #           EXPAND1  & mods -> perform necessary expansions on children first to ensure we don't end
+                    #                              up expanding inside our parents if (after expansion) we have more
+                    #                              than one child.
+                    subtree.extend(child.tail)
+                else:
+                    subtree.append(child)
+            if len(subtree) == 1 and (RuleMods.EXPAND in mods or RuleMods.EXPAND1 in mods):
+                # Self-expansion: only perform on EXPAND and EXPAND1 rules
+                p[0] = subtree[0]
+            else:
+                p[0] = self.tree_class(rule_name, subtree, skip_adjustments=True)
+        p_rule.__doc__ = rule_def
+        setattr(self, 'p_%s' % (rule_name,), types.MethodType(p_rule, self))
 
 
     @staticmethod
